#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <php.h>
#include <zend_exceptions.h>
#include <ext/standard/info.h>
#include <string.h>

#include <op5/license.h>
#include <op5/license_query.h>

#include "php_op5license.h"

ZEND_DECLARE_MODULE_GLOBALS(op5license)

PHP_MINIT_FUNCTION( op5license);
PHP_MSHUTDOWN_FUNCTION( op5license);
PHP_RINIT_FUNCTION( op5license);
PHP_RSHUTDOWN_FUNCTION( op5license);
PHP_MINFO_FUNCTION( op5license);
#ifdef COMPILE_DL_OP5LICENSE
ZEND_GET_MODULE(op5license)
#endif

PHP_INI_BEGIN() STD_PHP_INI_ENTRY("op5license.license_file_path", "/etc/op5license/op5license.lic", PHP_INI_ALL, OnUpdateString, license_file_path, zend_op5license_globals, op5license_globals)
PHP_INI_END()

/*******************************************************************************
 * op5license class-related stuff
 ******************************************************************************/

struct op5license_object {
	zend_object std;
	char *license_path;
	struct op5lic_t *lic;
	struct op5lic_metadata_t *metadata;
};

struct op5license_exception {
	zend_object std;
};

PHP_METHOD( op5license, __construct);
PHP_METHOD( op5license, run);
PHP_METHOD( op5license, get_metadata);
PHP_METHOD( op5license, __toString);

PHP_METHOD( op5license_exception, __toString);

void op5license_free_storage(void* object TSRMLS_DC);
void op5license_free_exception_storage(void * object TSRMLS_DC);
zend_object_value op5license_create_handler(zend_class_entry* type TSRMLS_DC);
zend_object_value op5license_create_exception(zend_class_entry* type TSRMLS_DC);

zend_object_handlers op5license_object_handlers;
zend_object_handlers op5license_exception_handlers;
zend_class_entry *op5license_ce;
zend_class_entry *op5license_ce_exception;

ZEND_BEGIN_ARG_INFO(arginfo_op5license_run, 0)
	ZEND_ARG_INFO(0, action)
	ZEND_ARG_INFO(0, environment) /* array */
	ZEND_ARG_INFO(1, messages) /* array */
	ZEND_ARG_INFO(1, metrics) /* array */
ZEND_END_ARG_INFO()

const zend_function_entry op5license_class_functions[] = {
		PHP_ME(op5license, __construct, NULL, ZEND_ACC_PUBLIC | ZEND_ACC_CTOR)
		PHP_ME(op5license, run, arginfo_op5license_run, ZEND_ACC_PUBLIC)
		PHP_ME(op5license, get_metadata, NULL, ZEND_ACC_PUBLIC)
		PHP_ME(op5license, __toString, NULL, ZEND_ACC_PUBLIC)
		{ NULL, NULL, NULL }
};

const zend_function_entry op5license_exception_functions[] = {
		PHP_ME(op5license_exception, __toString, NULL, ZEND_ACC_PUBLIC)
		{ NULL, NULL, NULL }
};

zend_module_entry op5license_module_entry = {
STANDARD_MODULE_HEADER, "op5license", op5license_class_functions,
PHP_MINIT(op5license),
PHP_MSHUTDOWN(op5license),
PHP_RINIT(op5license),
PHP_RSHUTDOWN(op5license),
PHP_MINFO(op5license),
PHP_OP5LICENSE_VERSION,
STANDARD_MODULE_PROPERTIES };


/*******************************************************************************
 * libop5license integration helpers
 ******************************************************************************/

/*******************************************************************************
 * Module related stuff
 ******************************************************************************/
PHP_MINIT_FUNCTION( op5license) {
	zend_class_entry ce;
	zend_class_entry ce_exception;
	REGISTER_INI_ENTRIES();

	INIT_CLASS_ENTRY(ce, "op5license", op5license_class_functions);
	op5license_ce = zend_register_internal_class(&ce TSRMLS_CC);
	op5license_ce->create_object = op5license_create_handler;
	memcpy(&op5license_object_handlers, zend_get_std_object_handlers(),
			sizeof(zend_object_handlers));
	op5license_object_handlers.clone_obj = NULL;

	INIT_CLASS_ENTRY(ce_exception, "Op5LicenseException", op5license_exception_functions);
	op5license_ce_exception = zend_register_internal_class_ex(&ce_exception, zend_exception_get_default(TSRMLS_C), NULL TSRMLS_CC);
	op5license_ce_exception->create_object = op5license_create_exception;
	memcpy(&op5license_exception_handlers, zend_get_std_object_handlers(),
			sizeof(zend_object_handlers));
	op5license_exception_handlers.clone_obj = NULL;
	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION( op5license) {
	UNREGISTER_INI_ENTRIES();
	return SUCCESS;
}

PHP_RINIT_FUNCTION( op5license) {
	return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION( op5license) {
	return SUCCESS;
}

/* {{{ PHP_MINFO(op5license) */
PHP_MINFO_FUNCTION( op5license) {
	php_info_print_table_start();
	php_info_print_table_header(2, "op5license support", "enabled");
	php_info_print_table_end();

	DISPLAY_INI_ENTRIES();
}
/* }}} */

/*******************************************************************************
 * op5license class methods
 ******************************************************************************/

void op5license_free_storage(void* object TSRMLS_DC) {
	struct op5license_object *obj = (struct op5license_object*) object;


	if (obj->lic != NULL)
		op5lic_destroy(obj->lic);

	if (obj->metadata != NULL)
		op5lic_metadata_destroy(obj->metadata);

	obj->lic = NULL;
	obj->metadata = NULL;
	zend_objects_free_object_storage(&obj->std);
}

void op5license_free_exception_storage(void *object TSRMLS_DC) {
	struct op5license_exception *obj = (struct op5license_exception*) object;
	zend_objects_free_object_storage(&obj->std);
}

void op5license_init_properties(zend_object *obj, zend_class_entry *type) {
#if PHP_VERSION_ID < 50399
	zval *tmp;
	zend_hash_copy(obj->properties, &type->default_properties,
			(copy_ctor_func_t) zval_add_ref, (void*) &tmp, sizeof(zval*));
#else
	object_properties_init(obj, type);
#endif
}

zend_object_value op5license_create_exception(zend_class_entry *type TSRMLS_DC) {
	zend_object_value retval;

	struct op5license_exception *exc = (struct op5license_exception*) emalloc(
			sizeof(struct op5license_exception));
	memset(exc, 0, sizeof(struct op5license_exception));
	zend_object_std_init(&exc->std, type TSRMLS_CC);
	op5license_init_properties(&exc->std, type);
	retval.handle  = zend_objects_store_put(exc, NULL,
			op5license_free_exception_storage, NULL TSRMLS_CC);
	retval.handlers = &op5license_exception_handlers;
	return retval;
}

zend_object_value op5license_create_handler(zend_class_entry* type TSRMLS_DC) {
	zend_object_value retval;

	struct op5license_object* obj = (struct op5license_object*) emalloc(
			sizeof(struct op5license_object));
	memset(obj, 0, sizeof(struct op5license_object));
	zend_object_std_init(&obj->std, type TSRMLS_CC);
	op5license_init_properties(&obj->std, type);

	retval.handle = zend_objects_store_put(obj, NULL, op5license_free_storage,
	NULL TSRMLS_CC);
	retval.handlers = &op5license_object_handlers;

	obj->lic = NULL;

	return retval;
}

PHP_METHOD( op5license, __construct) {
	struct op5license_object *obj;
	obj = (struct op5license_object *) zend_object_store_get_object(
	getThis() TSRMLS_CC);

	const char *license_path = OP5LICENSE_G(license_file_path);
	int license_path_length = strlen(license_path);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s", &license_path,
			&license_path_length) == FAILURE) {
		RETURN_NULL()
		;
	}

	obj->license_path = strndup(license_path, license_path_length);
	if (!obj->license_path) {
		RETURN_NULL();
	}
}

PHP_METHOD( op5license, __toString) {
	struct op5license_object *obj;
	obj = (struct op5license_object *) zend_object_store_get_object(
			getThis() TSRMLS_CC);

	RETURN_STRING(obj->license_path, 1);
}

PHP_METHOD( op5license_exception, __toString) {
	zval *message;
	message = zend_read_property(op5license_ce_exception, getThis(), "message",
			sizeof("message") - 1, 1 TSRMLS_CC);

	RETURN_STRING(Z_STRVAL_P(message), 1);
}


static struct op5lic_env_node_t *zval_to_op5lic_env_node(const char *key, zval *value) {
	struct op5lic_env_node_t *node = NULL;
	struct op5lic_env_node_t *subnode = NULL;
	zval **data;
	char *str_index = NULL;
	uint str_len;
	ulong num_index;
	HashPosition pos;
	HashTable *ht;
	switch (Z_TYPE_P(value)) {
		case IS_DOUBLE:
			node = op5lic_env_create_int(key, Z_DVAL_P(value));
			break;
		case IS_LONG:
		case IS_BOOL:
			node = op5lic_env_create_int(key, Z_LVAL_P(value));
			break;
		case IS_STRING:
			node = op5lic_env_create_string(key, Z_STRVAL_P(value));
			break;
		case IS_ARRAY:
			node = op5lic_env_create_list(key);
			ht = Z_ARRVAL_P(value);
			for (zend_hash_internal_pointer_reset_ex(ht, &pos);
					zend_hash_get_current_data_ex(ht, (void **) &data, &pos) == SUCCESS;
					zend_hash_move_forward_ex(ht, &pos)
					)
			{
				if (zend_hash_get_current_key_ex(ht, &str_index, &str_len, &num_index, 0, &pos) == HASH_KEY_IS_STRING) {
					subnode = zval_to_op5lic_env_node(str_index, *data);
					if (subnode)
						op5lic_env_list_add(node->value.val_env_list, subnode);
				}
			}
			break;
		default:
			zend_throw_exception_ex(op5license_ce_exception, 0 TSRMLS_DC, "Unhandled environment value type <%s>", zend_zval_type_name(value));
	}
	return node;
}

PHP_METHOD( op5license, run) {
	struct op5license_object *obj;

	char *action;
	int action_length;
	int result;
	zval *environment;
	zval *messages;
	zval *metrics;
	zval *metric;

	struct op5lic_query_response_t *result_obj = NULL;
	struct op5lic_query_t *lq = NULL;
	struct op5lic_env_node_t *env_node = NULL;
	struct op5lic_env_list_t *env = NULL;
	struct op5lic_error_t *err = NULL;
	struct string_list_t *message_list = NULL;
	struct metric_list_t *metric_list = NULL;
	struct metric_node_t *mnode = NULL;
	obj = (struct op5license_object *) zend_object_store_get_object(
	getThis() TSRMLS_CC);

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "saaa", &action,
			&action_length, &environment, &messages, &metrics) == FAILURE) {
		RETURN_NULL()
		;
	}

	if (obj->lic == NULL) {
		err = op5lic_error_create();
		obj->lic = op5lic_create(obj->license_path, err);
	}

	if (obj->lic == NULL) {
		result = (int) false;
		goto loser;
	}

	/*
	 * Build the query
	 */
	/*TODO: handle action_length */
	env_node = zval_to_op5lic_env_node("env", environment);
	env = op5lic_env_node_to_list(env_node);
	if (env == NULL) {
		env = op5lic_env_list_create();
	}

	lq = op5lic_query_create(obj->lic, action, env);
	if (lq == NULL) {
		result = (int) false;
		goto loser;
	}

	result_obj = op5lic_query_run(lq, err);
	if (result_obj == NULL) {
		result = (int) false;
		goto loser;
	}

	result = (int) op5lic_query_response_allow(result_obj);
	message_list = op5lic_query_response_message_list(result_obj);
	metric_list = op5lic_query_response_metric_list(result_obj);

	int i;
	for (i = 0; i < message_list->len; i++) {
		add_next_index_string(messages, message_list->strings[i], 1);
	}


	for (mnode = metric_list_head(metric_list); mnode != NULL; mnode =
			metric_node_next(mnode)) {
		MAKE_STD_ZVAL(metric);
		array_init(metric);
		add_next_index_long(metric, metric_node_value(mnode));
		add_next_index_long(metric, metric_node_min(mnode));
		add_next_index_long(metric, metric_node_max(mnode));
		add_assoc_zval(metrics, metric_node_id(mnode), metric);
	}

loser:
	op5lic_env_list_destroy(env);
	if (lq != NULL) {
		op5lic_query_destroy(lq);
	}
	if (result_obj != NULL) {
		op5lic_query_response_destroy(result_obj);
	}
	if (err != NULL && err->status != OP5LIC_STATUS_OK) {
		zend_throw_exception(op5license_ce_exception, err->description, 0 TSRMLS_CC);
	}
	op5lic_error_destroy(err);
	RETURN_BOOL(result);
}

void metadata_to_array(struct op5lic_metadata_t *md, zval **arr_out) {
	struct op5lic_metadata_iter_t iter;
	void *value;
	char *key;
	enum MetadataFieldType vtype;

	op5lic_metadata_iter_init(&iter, md);
	while (op5lic_metadata_iter_next(&iter, &key, &value, &vtype)) {
		zval *mdval;
		switch (vtype) {
		case MD_TYPE_TABLE:
			MAKE_STD_ZVAL(mdval);
			array_init(mdval);
			metadata_to_array((struct op5lic_metadata_t *)value, &mdval);
			add_assoc_zval(*arr_out, key, mdval);
			break;
		case MD_TYPE_STRING:
			add_assoc_string(*arr_out, key, (char *) value, 1);
			break;
		}
	}
}

PHP_METHOD( op5license, get_metadata) {
	struct op5license_object *obj;
	obj = (struct op5license_object *) zend_object_store_get_object(
	getThis() TSRMLS_CC);
	struct op5lic_error_t *err = NULL;
	if (obj->lic == NULL) {
		err = op5lic_error_create();
		obj->lic = op5lic_create(obj->license_path, err);
	}

	if (obj->lic == NULL) {
		goto loser;
	}

	array_init(return_value);
	if (!obj->metadata) {
		obj->metadata = op5lic_query_metadata(obj->lic, err);
	}

	if (obj->metadata) {
		metadata_to_array(obj->metadata, &return_value);
	}
loser:
	if (err != NULL && err->status != OP5LIC_STATUS_OK && err->status != OP5LIC_STATUS_MISSING_METADATA) {
		zend_throw_exception(op5license_ce_exception, err->description, 0 TSRMLS_CC);
	}
	op5lic_error_destroy(err);
}
